import math

DEBUG = True

# This screams for an auxiliary module.
OCTET_LENGTH = 4
MAX_DIGIT = 255

# Base of the hexadecimal number system.
HEX_BASE = 16

# TODO: Maybe access the resolution of each
# vertical and horizontal pixel and
# divide them by 2. This apex is fixed for now.
# Used for mirroring the pixels.
IMAGE_APEX = 8

FIXED_PIXEL = 'o'
OPTIONAL_PIXEL = '+'

# TODO: Split into two methods - Finding fixed pixels and optional pixels?
# Less efficient but does not damage my brain. Or generate a tupel of two lists in a seperate method
# and handle them differently.
def parse_pagan_file(filename, hashcode, sym=False, invert=False):
    fd = open(filename, 'r')

    drawmap = []
    optmap = []

    i = 0
    for line in fd:
        j = 0
        for char in line:
            if (char == FIXED_PIXEL):
                drawmap.append((i - 1, j - 1))
            if (char == OPTIONAL_PIXEL):
                optmap.append((i - 1, j - 1))
            j += 1
        i += 1

    fd.close()

    # TODO: Decide for each optional pixel to be displayed
    # in respect of the optmap-size and an the third ip octet.
    # Merge the result with drawmap.
    # print (optmap)
    #print (len(optmap))

    extmap = decideoptionalpixels(optmap, hashcode)

    drawmap += extmap

    if sym:
        drawmap = enforce_vertical_symmetry(drawmap)
    elif invert:
        drawmap = invert_vertical(drawmap)

    return sorted(drawmap)


# Use one of the octets to determine if a pixel is shown or not.
def decideoptionalpixels(optmap, hashcode):
    resmap = []

    # Start with the last character of the hash.
    opt_control = hashcode

    for pixel in optmap:

        control_dec = int(opt_control[-1], HEX_BASE)
        # control_dec = int('7', HEX_BASE)
        opt_control = opt_control[:-1]
        if ((control_dec % 2) == 0):
            resmap.append(pixel)

    return resmap


def enforce_vertical_symmetry(pixmap):
    '''Enforces vertical symmetry of the pixelmap.
    Returns a pixelmap with all pixels mirrored in the middle.
    The initial ones still remain.'''
    mirror = []
    for item in pixmap:
        y = item[0]
        x = item[1]

        if x <= IMAGE_APEX:
            diff_x = diff(x, IMAGE_APEX)
            mirror.append((y, x + (2 * diff_x) - 1))

        if x > IMAGE_APEX:
            diff_x = diff(x, IMAGE_APEX)
            mirror.append((y, x - (2 * diff_x) - 1))

    return mirror + pixmap


def invert_vertical(pixmap):
    '''Inverts all virtual pixels vertically. This results in
    a different pixelmap, where all pixels are mirrored on the
    vertical middle.'''
    mirror = []
    for item in pixmap:
        y = item[0]
        x = item[1]

        if x <= IMAGE_APEX:
            diff_x = diff(x, IMAGE_APEX)
            mirror.append((y, x + (2 * diff_x) - 1))

        if x > IMAGE_APEX:
            diff_x = diff(x, IMAGE_APEX)
            mirror.append((y, x - (2 * diff_x) - 1))

    return mirror


def diff(a, b):
    '''Returns the difference between two values.'''
    return int(math.fabs(a - b))


import math

# Number of colors.
COLOR_QUANTITY = 8

# Length of a color in hex.
HEX_COLOR_LEN = 6

# Base of the hexadecimal number system.
HEX_BASE = 16

# To generate unique colors, hashes need to
# contain at least this many characters.
MINIMUM_HASH_LEN = COLOR_QUANTITY * HEX_COLOR_LEN

# The first and second six characters of a hash
# determine the avatars overall aspect.
ASPECT_CONTROL_LEN = 6

# Decimal representation of hexadecimal 'ffffff'
# as the maximum value for aspect decisions.
MAX_DECISION_VALUE = 16777215

# Set True to generate debug output in this module.
DEBUG = True

WEAPONSTYLES = ['ONEHANDED_ONEHANDED',
                'SHIELD_ONEHANDED',
                'ONEHANDED',
                'TWOHANDED']

TWOHANDED_WEAPONS = ['GREATSWORD',
                     'BIGHAMMER',
                     'GREATMACE',
                     'GREATAXE',
                     'WAND']

ONEHANDED_WEAPONS = ['SWORD',
                     'HAMMER',
                     'AXE',
                     'FLAIL',
                     'MACE',
                     'DAGGER']

SHIELDS = ['LONGSHIELD',
           'ROUNDSHIELD',
           'BUCKLER',
           'SHIELD']

# The different aspects an avatar can take.
ASPECTSTYLES = [['HAIR'],
                ['HAIR', 'PANTS', 'TOP'],
                ['HAIR', 'PANTS'],
                ['HAIR', 'BOOTS', 'TOP'],
                ['HAIR', 'BOOTS'],
                ['HAIR', 'TOP'],
                ['HAIR', 'PANTS', 'BOOTS'],
                ['HAIR', 'PANTS', 'BOOTS', 'TOP'],
                ['PANTS', 'BOOTS', 'TOP'],
                ['PANTS', 'BOOTS'],
                ['PANTS', 'TOP'],
                ['PANTS'],
                ['BOOTS', 'TOP'],
                ['BOOTS'],
                ['TOP'],
                []]


def init_weapon_list():
    """Initialize the possible weapon
    combinations."""
    twohand = []
    for item in TWOHANDED_WEAPONS:
        twohand.append([item])

    onehand = []
    for item in ONEHANDED_WEAPONS:
        onehand.append([item])

    shield = SHIELDS

    dualwield_variations = []
    weapon_shield_variations = []

    for item in ONEHANDED_WEAPONS:
        for i in ONEHANDED_WEAPONS:
            dualwield_variations.append([item, i])
        for j in shield:
            weapon_shield_variations.append([j, item])

    return twohand + onehand + dualwield_variations + weapon_shield_variations


def grind_hash_for_colors(hashcode):
    """Extracts information from the hashcode
    to generate different colors. Returns a
    list of colors in (r,g,b) tupels."""

    # When using smaller hash algorithms like MD5 or SHA1,
    # the number of bits does not provide enough information
    # to generate unique colors. Instead the hash is internally
    # appended by itself to fit the MINIMUM_HASH_LEN.
    # This leads to smaller hashes displaying less color
    # variatons, depicting the insecurity of small hashes.
    while (len(hashcode) < MINIMUM_HASH_LEN):
        chardiff = diff(len(hashcode), MINIMUM_HASH_LEN)
        if DEBUG:
            print ("Hashcode: %r with length: %d is too small. Appending difference." % (hashcode, len(hashcode)))
        hashcode += hashcode[:chardiff]
        if DEBUG:
            print ("Hash is now: %r with length: %d" % (hashcode, len(hashcode)))

    hashparts = split_sequence(hashcode, HEX_COLOR_LEN)

    colors = []

    for i in range(COLOR_QUANTITY):
        colors.append(hex2rgb(hashparts[i]))
    if DEBUG:
        print ("Generated colors: %r" % colors)
    return colors


def split_sequence(seq, n):
    """Generates tokens of length n from a sequence.
    The last token may be of smaller length."""
    tokens = []
    while seq:
        tokens.append(seq[:n])
        seq = seq[n:]
    return tokens


# Grinds the ip address for an aspect style to draw on the pixelmap.
def grind_hash_for_aspect(hashcode):
    # aspect_control = split_sequence(hashcode, ASPECT_CONTROL_CHARS)
    aspect_control = hashcode[:ASPECT_CONTROL_LEN]
    #print aspect_control
    hash_dec_value = int(aspect_control, HEX_BASE)
    decision = map_decision(MAX_DECISION_VALUE, len(ASPECTSTYLES), hash_dec_value)
    if DEBUG:
        print ("Aspect decision: %r" %decision)
    return choose_aspect(decision)


def grind_hash_for_weapon(hashcode):
    """ Grinds the given hashcode for a weapon to draw on
    the pixelmap. Utilizes the second six characters from the
    hashcode."""

    weaponlist = init_weapon_list()
    # The second six characters of the hash
    # control the weapon decision.
    weapon_control = hashcode[ASPECT_CONTROL_LEN:(ASPECT_CONTROL_LEN * 2)]

    # Decimal value of the hash chunk to map.
    hash_dec_value = int(weapon_control, HEX_BASE)
    decision = map_decision(MAX_DECISION_VALUE, len(weaponlist), hash_dec_value)

    return choose_weapon(decision, weaponlist)


def choose_weapon(decision, weapons):
    """Chooses a weapon from a given list
    based on the decision."""
    choice = []
    for i in range(len(weapons)):
        if (i < decision):
            choice = weapons[i]
    return choice


def choose_aspect(decision):
    """Chooses a style from ASPECTSTYLES
    based on the decision."""
    choice = []
    for i in range(len(ASPECTSTYLES)):
        if (i < decision):
            choice = ASPECTSTYLES[i]
    return choice


def map_decision(max_digitsum, num_decisions, digitsum):
    """Maps the domain to a number of decisions."""
    return (num_decisions / (float(max_digitsum) + 1)) * (float(digitsum) + 1)


def hex2rgb(hexvalue):
    """Converts a given hex color to
    its respective rgb color."""

    # Make sure a possible '#' char is eliminated
    # before processing the color.
    if ('#' in hexvalue):
        hexcolor = hexvalue.replace('#', '')
    else:
        hexcolor = hexvalue

    # Hex colors have a fixed length of 6 characters excluding the '#'
    # TODO: Include custom exception here, even if it should never happen.
    if (len(hexcolor) != 6):
        print ("Unexpected length of hex color value.\nSix characters excluding \'#\' expected.")
        return 0

    # Convert each two characters of
    # the hex to an RGB color value.
    r = int(hexcolor[0:2], HEX_BASE)
    g = int(hexcolor[2:4], HEX_BASE)
    b = int(hexcolor[4:6], HEX_BASE)

    return r, g, b


def diff(a, b):
    """Returns the difference between two values."""
    return int(math.fabs(a - b))


if __name__ == "__main__":

    # Testing some hashes.
    hash1 = "0396233d5b28eded8e34c1bf9dc80fae34756743594b9e5ae67f4f7d124d2e3d"
    hash2 = "ef101b0bc42f41e23e325f3da71daeff43ff7df9d41ff268e53a06c767de8487"
    hash3 = "ca4da36c48be1c0b87a7d575c73f6e42"

    h1 = grind_hash_for_colors(hash1)
    h2 = grind_hash_for_colors(hash2)
    h3 = grind_hash_for_colors(hash3)

    hex2rgb('#ffffff')
    hex2rgb('#ffff00')
    hex2rgb('#f5f5f5')


from pagan import Avatar

MD5 = pagan.generator.HASH_MD5
SHA1 = pagan.generator.HASH_SHA1
SHA224 = pagan.generator.HASH_SHA224
SHA256 = pagan.generator.HASH_SHA256
SHA384 = pagan.generator.HASH_SHA384
SHA512 = pagan.generator.HASH_SHA512


from setuptools import setup

setup(
    name='pagan',
    packages=['pagan'],
    use_2to3=True,
    include_package_data=True,
    version='0.3.2',
    url='https://github.com/daboth/pagan',
    download_url='https://github.com/daboth/pagan/tarball/0.3.2',
    license='GPL',
    description='python avatar generator for absolute nerds',
    long_description=open('README.md').read(),
    author='David Bothe',
    author_email='davbothe@googlemail.com',
    keywords=['avatar', 'identicon', 'generator'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    install_requires=['Pillow']

)


import generator
import os


class Avatar():

    # Default output path is in the current working directory.
    DEFAULT_OUTPUT_PATH = ('%s/output/' % os.getcwd())
    # Default filename.
    DEFAULT_FILENAME = ('pagan')

    DEFAULT_HASHFUN = generator.HASH_MD5

    def __init__(self, inpt, hashfun=DEFAULT_HASHFUN):
        """Initialize the avatar and creates the image."""
        self.img = self.__create_image(inpt, hashfun)

    def __create_image(self, inpt, hashfun):
        """Creates the avatar based on the input and
        the chosen hash function."""
        if hashfun not in generator.HASHES.keys():
            print ("Unknown or unsupported hash function. Using default: %s" % self.DEFAULT_HASHFUN)
            algo = self.DEFAULT_HASHFUN
        else:
            algo = hashfun
        return generator.generate(inpt, algo)

    def show(self):
        """Shows a preview of the avatar in an external
        image viewer."""
        self.img.show()

    def change(self, inpt, hashfun=DEFAULT_HASHFUN):
        """Change the avatar by providing a new input.
        Uses the standard hash function if no one is given."""
        self.img = self.__create_image(inpt, hashfun)

    def save(self, path=DEFAULT_OUTPUT_PATH, filename=DEFAULT_FILENAME):
        """Saves a avatar under the given output path to
        a given filename. The file ending ".png" is appended
        automatically. If the path does not exist, it will be
        created. When no parameters are omitted, a default path
        and/or filename will be used."""

        # Creates a path when it does not exist.
        if not os.path.exists(path):
            os.makedirs(path)

        # Cut the .png file ending if one was omitted.
        if filename[-4:] == ".png":
            filename = filename[:-4]

        # Saves the image under the given filepath.
        filepath = ("%s%s.png" % (path, filename))
        print ("Saving: %s" % filepath)
        self.img.save(filepath, 'PNG', transparency=0)


from PIL import Image, ImageDraw
import hashgrinder
import pgnreader
import hashlib
import os

# Set True to generate debug output in this module.
DEBUG = False

# Default output path is the current working directory.
OUTPUT_PATH = ('%s%soutput%s' % (os.getcwd(), os.sep, os.sep))

#Directory of the installed module.
PACKAGE_DIR = os.path.dirname(os.path.abspath(__file__))

# Actual Image size in pixel is fixed in this version of pagan.
IMAGE_SIZE = (128, 128)

# Desired virtual resolution of the image output.
# Needs to be lesser or equal to the actual image size.
# (Fixed size for this pagan version.)
VIRTUAL_RESOLUTION = (16, 16)

# Size variations only allowed on powers of two,
# starting with 16 and ending at 2048. Not used yet.
# Current version only supports a fixed size of 16x16
# virtual pixels.
ALLOWED_RES = [16, 32, 64, 128, 256, 512, 1024, 2048]

# Image boundaries. Image starts at (0,0)
MAX_X = IMAGE_SIZE[0] - 1
MAX_Y = IMAGE_SIZE[1] - 1

# Imagemode ist fixed to RGBA for creating PNG-Files.
IMAGE_MODE = 'RGBA'

# All images are transparent.
BACKGROUND_COLOR = (0, 0, 0, 0)

# Tailor the path to the pagan files os independently.
FILE_BODY = ('%s%spgn%sBODY.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_BOOTS = ('%s%spgn%sBOOTS.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_SUBFIELD = ('%s%spgn%sSUBFIELD.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_MIN_SUBFIELD = ('%s%spgn%sMIN_SUBFIELD.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_TORSO = ('%s%spgn%sTORSO.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_HAIR = ('%s%spgn%sHAIR.pgn' % (PACKAGE_DIR, os.sep, os.sep))
FILE_SHIELD_DECO = ('%s%spgn%sSHIELD_DECO.pgn' % (PACKAGE_DIR, os.sep, os.sep))


# Out of the box supported
# HASH algorithms from pythons hashlib.
HASH_MD5 = 0
HASH_SHA1 = 1
HASH_SHA224 = 2
HASH_SHA256 = 3
HASH_SHA384 = 4
HASH_SHA512 = 5

# String representation of the hashes for debugging.
HASHES = {0 : "MD5", 1 : "SHA1", 2 : "SHA224",
          3 : "SHA256", 4 : "SHA384", 5 : "SHA512"}


im = Image.new(IMAGE_MODE, IMAGE_SIZE, BACKGROUND_COLOR)

verticalpixels = []
horizontalpixels = []

# The pixelmap contains the native pixels scaled to virtual pixels
# in the desired resolution. Virtual pixels have a start- and endpoint
# based on the native pixelgrid. Each virtual pixel is
# formatted as followed:
# (startpoint , endpoint , RGBA Color)
pixelmap = []

# Size of a single virtual pixel mapped to the real image size.
dotsize = (IMAGE_SIZE[0] / VIRTUAL_RESOLUTION[0], IMAGE_SIZE[1] / VIRTUAL_RESOLUTION[1])


def hash_input(inpt, algo=HASH_SHA256):
    """Generates a hash from a given String
    with a specified algorithm and returns the
    hash in hexadecimal form. Default: sha256."""
    if (algo == HASH_MD5):
        hashcode = hashlib.md5()
    elif (algo == HASH_SHA1):
        hashcode = hashlib.sha1()
    elif (algo == HASH_SHA224):
        hashcode = hashlib.sha224()
    elif (algo == HASH_SHA256):
        hashcode = hashlib.sha256()
    elif (algo == HASH_SHA384):
        hashcode = hashlib.sha384()
    elif (algo == HASH_SHA512):
        hashcode = hashlib.sha512()

    hashcode.update(inpt)
    hexhash = hashcode.hexdigest()
    return hexhash


def create_shield_deco_layer(weapons, ip):
    '''Reads the SHIELD_DECO.pgn file and creates
    the shield decal layer.'''
    layer = []
    if weapons[0] in hashgrinder.SHIELDS:
        layer = pgnreader.parse_pagan_file(FILE_SHIELD_DECO, ip, invert=False, sym=False)
    return layer


def create_hair_layer(aspect, ip):
    '''Reads the HAIR.pgn file and creates
    the hair layer.'''
    layer = []
    if 'HAIR' in aspect:
        layer = pgnreader.parse_pagan_file(FILE_HAIR, ip, invert=False, sym=True)
    return layer


def create_torso_layer(aspect, ip):
    '''Reads the TORSO.pgn file and creates
    the torso layer.'''
    layer = []
    if 'TOP' in aspect:
        layer = pgnreader.parse_pagan_file(FILE_TORSO, ip, invert=False, sym=True)
    return layer


def create_subfield_layer(aspect, ip):
    '''Reads the SUBFIELD.pgn file and creates
    the subfield layer.'''
    layer = []
    if 'PANTS' in aspect:
        layer = pgnreader.parse_pagan_file(FILE_SUBFIELD, ip, invert=False, sym=True)
    else:
        layer = pgnreader.parse_pagan_file(FILE_MIN_SUBFIELD, ip, invert=False, sym=True)

    return layer


def create_boots_layer(aspect, ip):
    '''Reads the BOOTS.pgn file and creates
    the boots layer.'''
    layer = []
    if 'BOOTS' in aspect:
        layer = pgnreader.parse_pagan_file(FILE_BOOTS, ip, invert=False, sym=True)
    return layer


def create_shield_layer(shield, hashcode):
    """Creates the layer for shields."""
    return pgnreader.parse_pagan_file(('%s%spgn%s' % (PACKAGE_DIR, os.sep, os.sep)) + shield + '.pgn', hashcode, sym=False, invert=False)


def create_weapon_layer(weapon, hashcode, isSecond=False):
    """Creates the layer for weapons."""
    return pgnreader.parse_pagan_file(('%s%spgn%s' % (PACKAGE_DIR, os.sep, os.sep)) + weapon + '.pgn', hashcode, sym=False, invert=isSecond)


def scale_pixels(color, layer):
    """Scales the pixel to the virtual pixelmap."""
    pixelmap = []

    # Scaling the pixel offsets.
    for pix_x in range(MAX_X + 1):
        for pix_y in range(MAX_Y + 1):

            # Horizontal pixels
            y1 = pix_y * dotsize[0]
            x1 = pix_x * dotsize[1]

            # Vertical pixels
            y2 = pix_y * dotsize[0] + (dotsize[0] - 1)
            x2 = pix_x * dotsize[1] + (dotsize[1] - 1)

            if (y1 <= MAX_Y) and (y2 <= MAX_Y):
                if (x1 <= MAX_X) and (x2 <= MAX_X):
                    if (pix_x, pix_y) in layer:
                        pixelmap.append([(y1, x1), (y2, x2), color])
    return pixelmap


def draw_image(pixelmap, img):
    '''Draws the image based on the given pixelmap.'''
    for item in pixelmap:
        color = item[2]

        # Access the rectangle edges.
        pixelbox = (item[0][0], item[0][1], item[1][0], item[1][1])
        draw = ImageDraw.Draw(img)
        draw.rectangle(pixelbox, fill=color)


def setup_pixelmap(hashcode):
    """Creates and combines all required layers to
    build a pixelmap for creating the virtual
    pixels."""

    # Color distribution.
    # colors = hashgrinder.grindIpForColors(ip)
    colors = hashgrinder.grind_hash_for_colors(hashcode)
    color_body = colors[0]
    color_subfield = colors[1]
    color_weapon_a = colors[2]
    color_weapon_b = colors[3]
    color_shield_deco = colors[4]
    color_boots = colors[5]
    color_hair = colors[6]
    color_top = colors[7]

    # Grinds for the aspect.
    aspect = hashgrinder.grind_hash_for_aspect(hashcode)


    #Determine weapons of the avatar.
    weapons = hashgrinder.grind_hash_for_weapon(hashcode)



    if DEBUG:
        print ("Current aspect: %r" % aspect)
        print ("Current weapons: %r" % weapons)

    # There is just one body template. The optional pixels need to be mirrored so
    # the body layout will be symmetric to avoid uncanny looks.
    layer_body = pgnreader.parse_pagan_file(FILE_BODY, hashcode, invert=False, sym=True)

    layer_hair = create_hair_layer(aspect, hashcode)
    layer_boots = create_boots_layer(aspect, hashcode)
    layer_torso = create_torso_layer(aspect, hashcode)

    has_shield = (weapons[0] in hashgrinder.SHIELDS)

    if has_shield:
        layer_weapon_a = create_shield_layer(weapons[0], hashcode)
        layer_weapon_b = create_weapon_layer(weapons[1], hashcode)
    else:
        layer_weapon_a = create_weapon_layer(weapons[0], hashcode)
        if (len(weapons) == 2):
            layer_weapon_b = create_weapon_layer(weapons[1], hashcode, True)

    layer_subfield = create_subfield_layer(aspect, hashcode)
    layer_deco = create_shield_deco_layer(weapons, hashcode)

    pixelmap = scale_pixels(color_body, layer_body)
    pixelmap += scale_pixels(color_top, layer_torso)
    pixelmap += scale_pixels(color_hair, layer_hair)
    pixelmap += scale_pixels(color_subfield, layer_subfield)
    pixelmap += scale_pixels(color_boots, layer_boots)
    pixelmap += scale_pixels(color_weapon_a, layer_weapon_a)
    if (len(weapons) == 2):
        pixelmap += scale_pixels(color_weapon_b, layer_weapon_b)
    pixelmap += scale_pixels(color_shield_deco, layer_deco)

    return pixelmap


def generate(str, alg):
    """Generates an PIL image avatar based on the given
    input String. Acts as the main accessor to pagan."""
    img = Image.new(IMAGE_MODE, IMAGE_SIZE, BACKGROUND_COLOR)
    hashcode = hash_input(str, alg)
    pixelmap = setup_pixelmap(hashcode)
    draw_image(pixelmap, img)
    return img